import json
from logging import getLogger
from typing import Any, Dict

logger = getLogger(__name__)


RESOURCE_TYPES = ["apply_start", "apply_progress", "apply_complete", "planned_change"]

RESOURCE_STATUSES = {
    "planned_change": "planned",
    "apply_start": "starting",
    "apply_progress": "in_progress",
    "apply_complete": "finished",
}


class TerraformApplyLog:

    resources: Dict[str, Any] = {}

    def add_lines(self, log_lines: str) -> None:
        for log_line in log_lines.split("\n"):
            if len(log_line) < 1:
                continue
            self.add_line(log_line)

    def add_line(self, log_line: str) -> None:

        log = json.loads(log_line)

        # apply_start: {"@level":"info","@message":"module.vpc.module.nat_gateway[0].aws_route.route_to_nat[1]: Destroying... [id=r-rtb-06df695578f26432f1080289494]","@module":"terraform.ui","@timestamp":"2022-08-11T10:17:40.692416-05:00","hook":{"resource":{"addr":"module.vpc.module.nat_gateway[0].aws_route.route_to_nat[1]","module":"module.vpc.module.nat_gateway[0]","resource":"aws_route.route_to_nat[1]","implied_provider":"aws","resource_type":"aws_route","resource_name":"route_to_nat","resource_key":1},"action":"delete","id_key":"id","id_value":"r-rtb-06df695578f26432f1080289494"},"type":"apply_start"}
        # apply_progress: {"@level":"info","@message":"module.vpc.aws_internet_gateway_attachment.main: Still destroying... [10s elapsed]","@module":"terraform.ui","@timestamp":"2022-08-11T10:17:50.696783-05:00","hook":{"resource":{"addr":"module.vpc.aws_internet_gateway_attachment.main","module":"module.vpc","resource":"aws_internet_gateway_attachment.main","implied_provider":"aws","resource_type":"aws_internet_gateway_attachment","resource_name":"main","resource_key":null},"action":"delete","elapsed_seconds":10},"type":"apply_progress"}
        # apply_progress: {"@level":"info","@message":"module.vpc.module.nat_gateway[0].aws_nat_gateway.main[0]: Still creating... [1m40s elapsed]","@module":"terraform.ui","@timestamp":"2022-08-11T09:56:00.104645-05:00","hook":{"resource":{"addr":"module.vpc.module.nat_gateway[0].aws_nat_gateway.main[0]","module":"module.vpc.module.nat_gateway[0]","resource":"aws_nat_gateway.main[0]","implied_provider":"aws","resource_type":"aws_nat_gateway","resource_name":"main","resource_key":0},"action":"create","elapsed_seconds":100},"type":"apply_progress"}
        # apply_complete: {"@level":"info","@message":"module.vpc.module.nat_gateway[0].aws_route.route_to_nat[2]: Destruction complete after 0s","@module":"terraform.ui","@timestamp":"2022-08-11T10:17:41.454174-05:00","hook":{"resource":{"addr":"module.vpc.module.nat_gateway[0].aws_route.route_to_nat[2]","module":"module.vpc.module.nat_gateway[0]","resource":"aws_route.route_to_nat[2]","implied_provider":"aws","resource_type":"aws_route","resource_name":"route_to_nat","resource_key":2},"action":"delete","elapsed_seconds":0},"type":"apply_complete"}
        # planned_change: {"@level":"info","@message":"module.vpc.module.availability_zones[2].aws_subnet.public: Plan to create","@module":"terraform.ui","@timestamp":"2022-08-11T09:54:03.688924-05:00","change":{"resource":{"addr":"module.vpc.module.availability_zones[2].aws_subnet.public","module":"module.vpc.module.availability_zones[2]","resource":"aws_subnet.public","implied_provider":"aws","resource_type":"aws_subnet","resource_name":"public","resource_key":null},"action":"create"},"type":"planned_change"}
        if log["type"] in RESOURCE_TYPES:
            self.process_resource(log)
            return

        # {"@level":"info","@message":"Apply complete! Resources: 32 added, 0 changed, 0 destroyed.","@module":"terraform.ui","@timestamp":"2022-08-11T09:56:06.970120-05:00","changes":{"add":32,"change":0,"remove":0,"operation":"apply"},"type":"change_summary"}
        if log["type"] == "change_summary":
            self.changes = log["changes"]
            return

        # {"@level":"info","@message":"Terraform 1.1.7","@module":"terraform.ui","@timestamp":"2022-08-11T09:58:58.654191-05:00","terraform":"1.1.7","type":"version","ui":"1.0"}
        if log["type"] == "version":
            self.terraform_version = log["ui"]
            return

        # {"@level":"info","@message":"Outputs: 1","@module":"terraform.ui","@timestamp":"2022-08-11T09:59:00.389242-05:00","outputs":{"vpc":{"sensitive":false,"type":["object",{"arn":"string","assign_generated_ipv6_cidr_block":"bool","cidr_block":"string","default_network_acl_id":"string","default_route_table_id":"string","default_security_group_id":"string","dhcp_options_id":"string","enable_classiclink":"bool","enable_classiclink_dns_support":"bool","enable_dns_hostnames":"bool","enable_dns_support":"bool","id":"string","instance_tenancy":"string","ipv4_ipam_pool_id":"string","ipv4_netmask_length":"number","ipv6_association_id":"string","ipv6_cidr_block":"string","ipv6_cidr_block_network_border_group":"string","ipv6_ipam_pool_id":"string","ipv6_netmask_length":"number","main_route_table_id":"string","owner_id":"string","tags":["map","string"],"tags_all":["map","string"]}],"value":{"arn":"arn:aws:ec2:us-west-2:755842654594:vpc/vpc-010f9dfb4129daaca","assign_generated_ipv6_cidr_block":false,"cidr_block":"10.246.0.0/16","default_network_acl_id":"acl-0a5dd0254c60c41b0","default_route_table_id":"rtb-0b774b6dc44ee4bc8","default_security_group_id":"sg-0401f1cc7fdb04283","dhcp_options_id":"dopt-0e7128608c809b67b","enable_classiclink":false,"enable_classiclink_dns_support":false,"enable_dns_hostnames":true,"enable_dns_support":true,"id":"vpc-010f9dfb4129daaca","instance_tenancy":"default","ipv4_ipam_pool_id":null,"ipv4_netmask_length":null,"ipv6_association_id":"","ipv6_cidr_block":"","ipv6_cidr_block_network_border_group":"","ipv6_ipam_pool_id":"","ipv6_netmask_length":0,"main_route_table_id":"rtb-0b774b6dc44ee4bc8","owner_id":"755842654594","tags":{"Name":"test-246"},"tags_all":{"Name":"test-246"}}}},"type":"outputs"}
        if log["type"] == "outputs":
            self.outputs = log["outputs"]
            return

        logger.warning(f"Apply log includes unknown type: {log['type']}")

    def process_resource(self, resource_log) -> None:
        if resource_log["type"] not in RESOURCE_TYPES:
            raise ValueError(f"Unexpected log type passed: {resource_log['type']}")

        resource_key = "hook" if resource_log["type"].startswith("apply_") else "change"
        self.resources[resource_log[resource_key]["resource"]["addr"]] = {
            **resource_log[resource_key]["resource"],
            "action": resource_log[resource_key]["action"],
            "status": RESOURCE_STATUSES[resource_log["type"]],
        }



import subprocess
from pathlib import Path
from unittest.mock import patch

from glom import glom

from terrapyst import TerraformWorkspace
from terrapyst.mixins import TerraformRun
from terrapyst.plan import TerraformPlan

from .conftest import PROVIDER_CACHE


def test_version():
    workspace = TerraformWorkspace()
    assert isinstance(workspace.version, str), "Version should be a string."


def test_init(workspace_environment):
    workspace = TerraformWorkspace(workspace_environment)
    workspace.env["TF_PLUGIN_CACHE_DIR"] = PROVIDER_CACHE
    workspace.init()
    assert (Path(workspace_environment) / ".terraform").is_dir(), "Workspace has initialized."


def test_init_with_custom_backend_conf(workspace_environment):
    with patch.object(subprocess, "run", wraps=subprocess.run) as wrapped_subprocess_call:
        workspace = TerraformWorkspace(
            workspace_environment, backend_config_path=Path(workspace_environment) / "mock.tfbackend"
        )
        workspace.env["TF_PLUGIN_CACHE_DIR"] = PROVIDER_CACHE
        workspace.init()
        assert any(["-backend-config=" in arg for arg in wrapped_subprocess_call.call_args.args[0]])
        assert (Path(workspace_environment) / ".terraform").is_dir(), "Workspace has initialized."


def test_plan(workspace):
    results, plan = workspace.plan(error_function=print, output_function=print)
    assert results.successful, "Terraform plan succeeded."
    assert glom(plan, "raw_plan.variables.test_string.value") == "yes, this is a test"
    assert isinstance(plan, TerraformPlan), "Terraform plan returned on successful plan."


def test_apply_interaction(workspace):
    results, apply_log = workspace.apply(error_function=print, output_function=print)
    assert results.returncode == 1, "Terraform apply failed when interaction is required."


def test_apply_auto_approve(workspace):
    results, apply_log = workspace.apply(auto_approve=True, error_function=print, output_function=print)
    assert results.returncode == 0, "Terraform apply succeeded."


def test_destroy_interaction(workspace):
    results, apply_log = workspace.destroy(error_function=print, output_function=print)
    assert results.returncode == 1, "Terraform apply failed when interaction is required."


def test_destroy_auto_approve(workspace):
    results, apply_log = workspace.destroy(auto_approve=True, error_function=print, output_function=print)
    assert results.returncode == 0, "Terraform apply succeeded."



import json
import os
import shutil
import subprocess
import tempfile
from logging import getLogger
from pathlib import Path
from typing import Callable, Dict, Optional, Tuple, Union

from terrapyst.apply_log import TerraformApplyLog

from .exceptions import TerraformError, TerraformRuntimeError
from .mixins import ProcessResults, TerraformRun
from .plan import TerraformPlan

logger = getLogger(__name__)


class TerraformWorkspace(TerraformRun):
    env: Dict[str, str]

    def __init__(self, path=None, backend_config_path: Optional[Path] = None) -> None:
        self.backend_config_path = backend_config_path
        self.terraform_path = shutil.which("terraform")
        if not self.terraform_path:
            raise TerraformError("Terraform binary is missing from system.")

        results = subprocess.run(["terraform", "-version", "-json"], stdout=subprocess.PIPE)
        if results.returncode != 0:
            raise TerraformRuntimeError("Unable to get terraform version", results)

        version_data = json.loads(results.stdout.decode("utf-8"))
        self.version = version_data["terraform_version"]
        self.is_outdated = version_data["terraform_outdated"]

        if self.is_outdated:
            logger.warning(f"Terraform version {self.version} is out of date. Please consider updating!")

        self.platform = version_data["platform"]
        self.provider_selections = version_data["provider_selections"]

        self.cwd = path if path != None else Path(os.getcwd())
        self.env = {}

    def init(self) -> ProcessResults:
        run_command = [self.terraform_path, "init"]
        if self.backend_config_path:
            run_command.append(f"-backend-config={str(self.backend_config_path)}")
        return self._subprocess_run(run_command, raise_exception_on_failure=True)

    def validate(self) -> ProcessResults:
        return self._subprocess_run(["terraform", "validate", "-json"], raise_exception_on_failure=True)

    def plan(
        self, error_function=None, output_function=None, output_path=None, destroy=False
    ) -> Tuple[ProcessResults, Union[TerraformPlan, None]]:
        save_plan = True
        if not output_path:
            save_plan = False
            temp_dir = tempfile.mkdtemp()
            output_path = os.path.join(temp_dir, "terraform.plan")

        command = [
            self.terraform_path,
            "plan",
            "-json",
            "-detailed-exitcode",
            "-input=false",
            f"-out={output_path}",
        ]

        if destroy:
            command.append("-destroy")

        results = self._subprocess_stream(
            command,
            error_function=error_function,
            output_function=output_function,
        )

        if results.returncode != 1:
            results.successful = True
            plan = TerraformPlan(cwd=self.cwd, plan_path=output_path)
        else:
            results.successful = False
            plan = None

        if not save_plan:
            shutil.rmtree(temp_dir)

        return results, plan

    def apply(
        self,
        error_function: Callable = None,
        output_function: Callable = None,
        auto_approve: bool = False,
        plan_file: str = None,
    ) -> Tuple[ProcessResults, TerraformApplyLog]:

        command = [self.terraform_path, "apply", "-json"]

        if plan_file:
            command.append(plan_file)
        if auto_approve:
            command.append("-auto-approve")

        results = self._subprocess_stream(
            command,
            error_function=error_function,
            output_function=output_function,
        )

        apply_log = TerraformApplyLog()
        apply_log.add_lines(results.stdout)

        return results, apply_log

    def destroy(
        self, auto_approve: bool = False, error_function: Callable = None, output_function: Callable = None
    ) -> Tuple[ProcessResults, TerraformApplyLog]:
        terraform_command = [self.terraform_path, "destroy", "-json"]
        if auto_approve:
            terraform_command.append("-auto-approve")

        results = self._subprocess_stream(
            terraform_command,
            error_function=error_function,
            output_function=output_function,
        )

        apply_log = TerraformApplyLog()
        apply_log.add_lines(results.stdout)

        return results, apply_log

    def output(self) -> ProcessResults:
        return self._subprocess_run([self.terraform_path, "output", "-json"])

    def get(self, update: bool = False) -> ProcessResults:
        command = [self.terraform_path, "get"]
        if update:
            command.append("-update")
        return self._subprocess_run(command)



import shutil
import tempfile
from pathlib import Path

import pytest

from terrapyst import TerraformWorkspace

WORKSPACE_TEMPLATE = Path(__file__).parent.absolute() / "terraform" / "workspace"

# Speed up `init` functions by caching providers locally.
# This will be deleted between runs.
PROVIDER_CACHE = tempfile.mkdtemp()


@pytest.fixture
def workspace_environment(tmpdir):
    shutil.copytree(WORKSPACE_TEMPLATE, tmpdir, dirs_exist_ok=True)
    return tmpdir


@pytest.fixture
def workspace(workspace_environment):
    workspace = TerraformWorkspace(workspace_environment)
    workspace.env["TF_PLUGIN_CACHE_DIR"] = PROVIDER_CACHE
    workspace.init()
    return workspace


@pytest.fixture
def plan(workspace):
    results, plan = workspace.plan()
    return plan



class TerraformError(Exception):
    pass


class TerraformRuntimeError(TerraformError):
    def __init__(self, message, process_results) -> None:
        self.process_results = process_results
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message} \n Return Code: {self.process_results.returncode} \n\n STDOUT: \n {self.process_results.stdout} \n\n STDERR: \n {self.process_results.stderr}"


class TerraformVersionError(TerraformError):
    pass



import os
import subprocess
from logging import getLogger
from typing import IO, Any, Dict, cast

from .exceptions import TerraformRuntimeError

logger = getLogger(__name__)

DEFAULT_ENV_VARS = {"TF_IN_AUTOMATION": "1", "TF_INPUT": "0"}


class ProcessResults:
    def __init__(self, returncode, stdout, stderr) -> None:
        self.returncode = returncode
        self.successful = returncode == 0
        self.stdout = stdout
        self.stderr = stderr
        self.env: Dict[str, Any] = {}


class TerraformRun:
    cwd: str = os.getcwd()
    env: Dict[str, str] = {}

    def _subprocess_run(self, args, raise_exception_on_failure=False, **kwargs) -> ProcessResults:
        default_kwargs = {
            "cwd": self.cwd,
            "capture_output": True,
            "encoding": "utf-8",
            "timeout": None,
            "env": {**os.environ, **DEFAULT_ENV_VARS, **self.env},
        }
        pass_kwargs = {**default_kwargs, **kwargs}
        results = subprocess.run(args, **pass_kwargs)  # type: ignore

        ret_results = ProcessResults(results.returncode, results.stdout, results.stderr)

        if raise_exception_on_failure and not ret_results.successful:
            error_message = f"An error occurred while running command '{' '.join(args)}'\n"
            error_message += f"returncode: {ret_results.returncode}\n"
            error_message += f"stdout: {ret_results.stdout}\n"
            error_message += f"stderr: {ret_results.stderr}\n"
            raise TerraformRuntimeError(error_message, ret_results)

        return ret_results

    def _subprocess_stream(self, command, error_function=None, output_function=None, **kwargs) -> ProcessResults:
        logger.info(f"Running command '{command}'")
        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=self.cwd,
            env={**os.environ, **DEFAULT_ENV_VARS, **self.env},
            universal_newlines=True,
            encoding="utf-8",
            bufsize=1,
            **kwargs,
        )
        stdout = ""
        stderr = ""
        while True:
            had_output = False

            # Check for stdout changes.
            for output in cast(IO, process.stdout):
                stdout += output
                logger.info(output)
                if output_function:
                    output_function(output)

            # Check for stderr changes.
            for output in cast(IO, process.stderr):
                had_output = True
                stderr += output
                logger.warning(output)
                if error_function:
                    error_function(output)

            # Process is closed and we've read all of the outputs.
            if process.poll() is not None:
                if not had_output:
                    break

        # Match the return object of subprocess.run.
        return ProcessResults(returncode=process.poll(), stdout=stdout, stderr=stderr)



import json
import shutil
from logging import getLogger
from typing import Any, Dict

from .exceptions import TerraformRuntimeError, TerraformVersionError
from .mixins import TerraformRun

logger = getLogger(__name__)

MODIFICATION_ACTIONS = ["update"]
DELETION_ACTIONS = ["delete"]
CREATE_ACTIONS = ["create"]


class TerraformPlan(TerraformRun):
    changes: Dict[str, Any]
    env: Dict[str, str]

    def __init__(self, cwd, plan_path) -> None:
        # confirm file exists

        self.cwd = cwd
        self.env = {}

        terraform_path = shutil.which("terraform")
        command = [terraform_path, "show", "-json", plan_path]
        results = self._subprocess_run(command)

        if results.returncode != 0:
            raise TerraformRuntimeError("Terraform plan failed", results)

        plan_details = json.loads(results.stdout)
        self.raw_plan = plan_details
        self.terraform_version = plan_details["terraform_version"]
        self.format_version = plan_details["format_version"]

        if self.format_version[:1] != "1":
            raise TerraformVersionError(
                f"Expected semantic version equivalent of v1, instead found '{self.format_version}'"
            )

        self.deletions = 0
        self.creations = 0
        self.modifications = 0

        self.changes = {}
        self._parse_changes(plan_details.get("resource_changes", []))

    def _parse_changes(self, change_plan: list):
        for changeset in change_plan:
            change = TerraformChange(changeset)
            self.changes[changeset["address"]] = TerraformChange(changeset)
            if change.will_delete():
                self.deletions += 1
            if change.will_create():
                self.creations += 1
            if change.will_modify():
                self.modifications += 1


class TerraformChange:
    def __init__(self, changeset) -> None:
        self.address = changeset["address"]
        self.type = changeset["type"]
        self.actions = changeset["change"]["actions"]

    def will_delete(self):
        return len(list(set(self.actions) & set(DELETION_ACTIONS))) > 0

    def will_modify(self):
        return len(list(set(self.actions) & set(MODIFICATION_ACTIONS))) > 0

    def will_create(self):
        return len(list(set(self.actions) & set(CREATE_ACTIONS))) > 0



from pathlib import Path

import pytest

from terrapyst.apply_log import TerraformApplyLog

APPLY_LOG_FILE = Path(__file__).parent.absolute() / "terraform" / "applies" / "vpc.jsonl"
with open(APPLY_LOG_FILE) as f:
    APPLY_LOG_CONTENTS = f.read()


def test_read_log():
    apply_log = TerraformApplyLog()
    apply_log.add_lines(APPLY_LOG_CONTENTS)

    assert "module.vpc.module.nat_gateway[0].aws_route.route_to_nat[2]" in apply_log.resources
    assert apply_log.changes["operation"] == "apply"
    assert apply_log.changes["add"] == 32
    assert apply_log.changes["change"] == 0
    assert apply_log.changes["remove"] == 0
    assert "vpc" in apply_log.outputs


def test_warning(caplog):
    apply_log = TerraformApplyLog()
    apply_log.add_lines(APPLY_LOG_CONTENTS)
    apply_log.add_line('{"type":"fake"}')
    assert "Apply log includes unknown type:" in caplog.text



VERSION = "0.0.1"
from . import _version
from .apply_log import TerraformApplyLog
from .authentication import TerraformAuthentication
from .exceptions import TerraformError, TerraformRuntimeError, TerraformVersionError
from .plan import TerraformChange, TerraformPlan
from .workspace import TerraformWorkspace

__version__ = _version.get_versions()["version"]



from pathlib import Path

import pytest

PLAN_FILE = Path(__file__).parent.absolute() / "terraform" / "plans" / "test.plan"


@pytest.fixture
def plan(workspace):
    results, plan = workspace.plan()
    return plan


def test_load(plan):
    assert plan.format_version == "1.1"
    assert plan.deletions == 0, "Nothing to delete."
    assert plan.modifications == 0, "Nothing to modify."
    assert plan.creations == 7, "Creating 7 resources."


def test_reload(plan):
    plan._parse_changes([{"address": "1", "type": "test", "change": {"actions": ["delete"]}}])
    assert plan.format_version == "1.1"
    assert plan.deletions == 1, "Deleting 1 resources."
    assert plan.modifications == 0, "Nothing to modify."
    assert plan.creations == 7, "Creating 7 resources."



# This file is deprecated by python in favor of setup.cfg
# However, some tools haven't caught up so we leave this here for them.
# All dependency and build configuration comes from setup.cfg.
# Other configuration, such as for linters, is in pyproject.toml.

import setuptools
import versioneer

setuptools.setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)
